from ford.fortran_project import Project
from ford import DEFAULT_SETTINGS
from ford.graphs import graphviz_installed, GraphManager

from copy import deepcopy
from textwrap import dedent

import markdown
import pytest
from bs4 import BeautifulSoup


def create_project(settings: dict):
    md_ext = [
        "markdown.extensions.meta",
        "markdown.extensions.codehilite",
        "markdown.extensions.extra",
    ]
    md = markdown.Markdown(
        extensions=md_ext, output_format="html", extension_configs={}
    )

    project = Project(settings)
    project.markdown(md, "..")
    project.correlate()
    return project


@pytest.fixture(scope="module")
def make_project_graphs(tmp_path_factory):
    data = """\
    module a
    end module a

    module b
      use a
    end module b

    module c
      use b
      use iso_fortran_env
      use external_mod

      type base
      end type base

      type, extends(base) :: derived
      end type derived

      type, extends(derived) :: leaf
        type(external_type) :: component
      end type leaf

      type thing
        type(derived) :: part
      end type thing

      type alpha
      contains
        procedure :: five
        procedure :: six
      end type alpha

    contains
      subroutine one
      end subroutine one
      subroutine two
        call one
      end subroutine two
      subroutine five
      end subroutine five
      function six(this) result(res)
        real :: res
        res = 1
      end function six
    end module c

    submodule (c) c_submod
    end submodule c_submod

    submodule (c:c_submod) c_subsubmod
    end submodule c_subsubmod

    program foo
      use c
      call three
      real :: x
      type(alpha) :: y
      call y%five()
      x = y%six()
    contains
      subroutine three
        use external_mod
        call one
        call two
        call two
        call other_sub
      end subroutine three

      recursive subroutine four
        call four
      end subroutine four
    end program foo
    """

    src_dir = tmp_path_factory.getbasetemp() / "graphs" / "src"
    src_dir.mkdir(exist_ok=True, parents=True)
    full_filename = src_dir / "test.f90"
    with open(full_filename, "w") as f:
        f.write(dedent(data))

    settings = deepcopy(DEFAULT_SETTINGS)
    settings["src_dir"] = [src_dir]
    settings["graph"] = True
    project = create_project(settings)

    graphs = GraphManager(
        "", "", graphdir="", parentdir="..", coloured_edges=True, show_proc_parent=True
    )
    for entity_list in [
        project.types,
        project.procedures,
        project.submodprocedures,
        project.modules,
        project.submodules,
        project.programs,
        project.files,
        project.blockdata,
    ]:
        for item in entity_list:
            graphs.register(item)

    graphs.graph_all()
    graphs.output_graphs(0)
    return graphs


MOD_GRAPH_KEY = ["Module", "Submodule", "Subroutine", "Function", "Program"]
PROC_GRAPH_KEY = [
    "Subroutine",
    "Function",
    "Interface",
    "Unknown Procedure Type",
    "Program",
]
TYPE_GRAPH_KEY = ["Type"]


@pytest.mark.skipif(not graphviz_installed, reason="Requires graphviz")
@pytest.mark.parametrize(
    ("graph_name", "expected_nodes", "expected_edges", "expected_legend_nodes"),
    [
        (
            ["usegraph"],
            [
                "a",
                "b",
                "c",
                "c_submod",
                "c_subsubmod",
                "iso_fortran_env",
                "external_mod",
                "foo::three",
                "foo",
            ],
            [
                "module~b->module~a",
                "module~c->module~b",
                "module~c->iso_fortran_env",
                "module~c->external_mod",
                "module~c_submod->module~c",
                "module~c_subsubmod->module~c_submod",
                "proc~three->external_mod",
                "program~foo->module~c",
            ],
            MOD_GRAPH_KEY,
        ),
        (
            ["callgraph"],
            [
                "c::one",
                "foo::three",
                "c::two",
                "foo::four",
                "other_sub",
                "foo",
                "c::alpha%five",
                "c::alpha%six",
            ],
            [
                "proc~three->proc~one",
                "proc~three->proc~two",
                "proc~two->proc~one",
                "proc~three->other_sub",
                "program~foo->proc~three",
                "proc~four->proc~four",
                "program~foo->proc~five",
                "program~foo->proc~six",
            ],
            PROC_GRAPH_KEY,
        ),
        (
            ["typegraph"],
            ["base", "derived", "leaf", "external_type", "thing", "alpha"],
            [
                "type~leaf->type~derived",
                "type~derived->type~base",
                "type~leaf->external_type",
                "type~thing->type~derived",
            ],
            TYPE_GRAPH_KEY,
        ),
        (
            ["modules", "b", "usesgraph"],
            ["a", "b"],
            ["module~b->module~a"],
            MOD_GRAPH_KEY,
        ),
        (
            ["modules", "b", "usedbygraph"],
            ["b", "c", "c_submod", "c_subsubmod", "foo"],
            [
                "module~c->module~b",
                "module~c_submod->module~c",
                "module~c_subsubmod->module~c_submod",
                "program~foo->module~c",
            ],
            MOD_GRAPH_KEY,
        ),
        (
            ["types", "derived", "inhergraph"],
            ["base", "derived"],
            ["type~derived->type~base"],
            TYPE_GRAPH_KEY,
        ),
        (
            ["types", "derived", "inherbygraph"],
            ["derived", "leaf", "thing"],
            ["type~leaf->type~derived", "type~thing->type~derived"],
            TYPE_GRAPH_KEY,
        ),
        (
            ["procedures", "two", "callsgraph"],
            ["c::one", "c::two"],
            ["proc~two->proc~one"],
            PROC_GRAPH_KEY,
        ),
        (
            ["procedures", "two", "calledbygraph"],
            ["foo::three", "c::two", "foo"],
            ["proc~three->proc~two", "program~foo->proc~three"],
            PROC_GRAPH_KEY,
        ),
        (
            ["procedures", "three", "usesgraph"],
            ["external_mod", "foo::three"],
            ["proc~three->external_mod"],
            MOD_GRAPH_KEY,
        ),
    ],
)
def test_graphs(
    make_project_graphs,
    graph_name,
    expected_nodes,
    expected_edges,
    expected_legend_nodes,
):
    graphs = make_project_graphs

    # Get the graph we want from the graph manager. This might be a
    # top-level graph, showing an overview of the project, or it might
    # be a graph for a particular object. In the latter case, we need
    # to also find the object from the appropriate collection
    if len(graph_name) == 1:
        graph = getattr(graphs, graph_name[0])
    else:
        collection, obj_name, name = graph_name
        for graph in getattr(graphs, collection):
            if graph.name == obj_name:
                break
        graph = getattr(graph, name)

    soup = BeautifulSoup(str(graph), features="html.parser")
    # Get nodes and edges just in the graph, and not in the legend
    node_names = [s.find("text").text for s in soup.svg.find_all("g", class_="node")]
    edge_names = [s.title.text for s in soup.svg.find_all("g", class_="edge")]

    assert sorted(node_names) == sorted(expected_nodes)
    assert sorted(edge_names) == sorted(expected_edges)

    svgs = soup.find_all("svg")
    assert len(svgs) == 2, "Graph and legend"

    legend = svgs[1]
    legend_nodes = [s.title.text for s in legend.find_all("g", class_="node")]
    assert legend_nodes == expected_legend_nodes + ["This Page's Entity"]
    assert legend.find_all("g", class_="edge") == []


def test_graphs_as_table(tmp_path):
    data = """\
    program foo
    contains
      subroutine one
      end subroutine one

      subroutine a
        call one()
      end subroutine a

      subroutine b
        call one()
      end subroutine b

      subroutine c
        call one()
      end subroutine c

      subroutine d
        call one()
      end subroutine d

      subroutine e
        call one()
      end subroutine e

      subroutine f
        call one()
      end subroutine f

      subroutine g
        call one()
      end subroutine g

      subroutine h
        call one()
      end subroutine h
    end program foo
    """

    src_dir = tmp_path / "graphs" / "src"
    src_dir.mkdir(exist_ok=True, parents=True)
    full_filename = src_dir / "test.f90"
    with open(full_filename, "w") as f:
        f.write(dedent(data))

    settings = deepcopy(DEFAULT_SETTINGS)
    settings["src_dir"] = [src_dir]
    settings["graph"] = True
    settings["graph_maxnodes"] = 4
    project = create_project(settings)

    graphs = GraphManager(
        "", "", graphdir="", parentdir="..", coloured_edges=True, show_proc_parent=True
    )
    for entity_list in [
        project.procedures,
        project.programs,
    ]:
        for item in entity_list:
            graphs.register(item)

    graphs.graph_all()
    graphs.output_graphs(0)

    for graph in graphs.procedures:
        if graph.name == "one":
            break
    graph = graph.calledbygraph

    soup = BeautifulSoup(str(graph), features="html.parser")
    node_names = sorted([n.text for n in soup.table.find_all(class_="node")])
    num_arrows = len(soup.table.find_all(class_="triangle-right"))
    # These are the spacing 'w's
    num_ws = len(soup.table.find_all(class_="solidBottom"))

    expected_node_names = [
        "foo::a",
        "foo::b",
        "foo::c",
        "foo::d",
        "foo::e",
        "foo::f",
        "foo::g",
        "foo::h",
    ]

    assert node_names == expected_node_names
    assert num_arrows == len(expected_node_names)
    assert num_ws == len(expected_node_names)



import copy
import ford
import pathlib
import pytest
from textwrap import dedent
import subprocess

# Ford default src folder
DEFAULT_SRC = "src"


def copy_file(data: str, path: pathlib.Path, filename: str) -> pathlib.Path:
    """Write data to 'path/filename'"""
    path.mkdir(exist_ok=True)
    full_filename = path / filename
    with open(full_filename, "w") as f:
        f.write(dedent(data))
    return full_filename


@pytest.fixture
def copy_fortran_file(tmp_path):
    return lambda data, extension=".f90": copy_file(
        data, tmp_path / "src", f"test{extension}"
    )


@pytest.fixture
def copy_settings_file(tmp_path):
    return lambda data: copy_file(data, tmp_path, "test.md")


@pytest.fixture
def restore_macros():
    old_macros = copy.copy(ford.utils._MACRO_DICT)
    yield
    ford.utils._MACRO_DICT = copy.copy(old_macros)


@pytest.fixture
def restore_nameselector():
    yield
    ford.sourceform.namelist = ford.sourceform.NameSelector()


def gfortran_is_not_installed():
    """Returns False if gfortran is not (detectably) installed"""
    out = subprocess.run("command -v gfortran", shell=True, check=False)
    return out.returncode != 0



from setuptools import setup
import site
import sys

# https://github.com/pypa/pip/issues/7953
if len(sys.argv) >= 3 and sys.argv[1] == "develop":
    site.ENABLE_USER_SITE = "--user" in sys.argv[2:]

setup()



# -*- coding: utf-8 -*-
#
#  pagetree.py
#  This file is part of FORD.
#
#  Copyright 2015 Christopher MacMackin <cmacmackin@gmail.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

import os


class PageNode(object):
    """
    Object representing a page in a tree of pages and subpages.
    """

    base_url = ".."

    def __init__(self, md, path, proj_copy_subdir, parent):
        print("Reading page {}".format(os.path.relpath(path)))
        with open(path, "r") as page:
            text = md.reset().convert(page.read())

        if "title" in md.Meta:
            self.title = "\n".join(md.Meta["title"])
        else:
            raise Exception("Page {} has no title metadata".format(path))
        if "author" in md.Meta:
            self.author = "\n".join(md.Meta["author"])
        else:
            self.author = None
        if "date" in md.Meta:
            self.date = "\n".join(md.Meta["date"])
        else:
            self.date = None

        if "ordered_subpage" in md.Meta:
            # index.md is the main page, it should not be added by the user in the subpage lists.
            self.ordered_subpages = [
                x for x in md.Meta["ordered_subpage"] if x != "index.md"
            ]
        else:
            self.ordered_subpages = None

        # set list of directory names that are to be copied along without
        # containing an index.md itself.
        #   first priority is the copy_dir option in the *.md file
        #   if this option is not set in the file fall back to the global
        #   project settings
        if "copy_subdir" in md.Meta:
            self.copy_subdir = md.Meta["copy_subdir"]
        else:
            self.copy_subdir = proj_copy_subdir

        self.parent = parent
        self.contents = text
        self.subpages = []
        self.files = []
        if self.parent:
            self.hierarchy = self.parent.hierarchy + [self.parent]
        else:
            self.hierarchy = []

        self.filename = os.path.split(path)[1][:-3]
        if parent:
            self.topdir = parent.topdir
            self.location = os.path.relpath(os.path.split(path)[0], self.topdir)
            self.topnode = parent.topnode
        else:
            self.topdir = os.path.split(path)[0]
            self.location = ""
            self.topnode = self

    def __str__(self):
        # ~ urlstr = "<a href='{0}/page/{1}/{2}.html'>{3}</a>"
        urlstr = "<a href='{0}'>{1}</a>"
        url = urlstr.format(
            os.path.join(self.base_url, "page", self.location, self.filename + ".html"),
            self.title,
        )
        return url

    def __iter__(self):
        retlist = [self]
        for sp in self.subpages:
            retlist.extend(list(sp.__iter__()))
        return iter(retlist)


def get_page_tree(topdir, proj_copy_subdir, md, parent=None):
    # In python 3.6 or newer, the normal dict is guaranteed to be ordered.
    # However, to keep compatibility with older versions I use OrderedDict.
    # I will use this later to remove duplicates from a list in a short way.
    from collections import OrderedDict

    # look for files within topdir
    filelist = sorted(os.listdir(topdir))
    if "index.md" not in filelist:
        print(f"Warning: No index.md file in directory {topdir}")
        return None

    # process index.md
    try:
        node = PageNode(md, os.path.join(topdir, "index.md"), proj_copy_subdir, parent)
    except Exception as e:
        print(
            "Warning: Error parsing {}.\n\t{}".format(
                os.path.relpath(os.path.join(topdir, "index.md")), e.args[0]
            )
        )
        return None
    filelist.remove("index.md")

    if node.ordered_subpages:
        # Merge user given files and all files in folder, removing duplicates.
        mergedfilelist = list(OrderedDict.fromkeys(node.ordered_subpages + filelist))
    else:
        mergedfilelist = filelist

    for name in mergedfilelist:
        if name[0] != "." and name[-1] != "~":
            if not os.path.exists(os.path.join(topdir, name)):
                raise Exception("Requested page file {} does not exist.".format(name))
            elif os.path.isdir(os.path.join(topdir, name)):
                # recurse into subdirectories
                traversedir = True
                if parent is not None:
                    traversedir = name not in parent.copy_subdir
                if traversedir:
                    subnode = get_page_tree(
                        os.path.join(topdir, name), proj_copy_subdir, md, node
                    )
                    if subnode:
                        node.subpages.append(subnode)
            elif name[-3:] == ".md":
                # process subpages
                try:
                    node.subpages.append(
                        PageNode(md, os.path.join(topdir, name), proj_copy_subdir, node)
                    )
                except Exception as e:
                    print(
                        "Warning: Error parsing {}.\n\t{}".format(
                            os.path.relpath(os.path.join(topdir, name)), e.args[0]
                        )
                    )
                    continue
            else:
                node.files.append(name)
    return node


def set_base_url(url):
    PageNode.base_url = url



# -*- coding: utf-8 -*-
#
#  graphs.py
#  This file is part of FORD.
#
#  Copyright 2015 Christopher MacMackin <cmacmackin@gmail.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from __future__ import annotations

import colorsys
import copy
import itertools
import os
import pathlib
import re
from typing import Dict, Iterable, List, Optional, Set, Tuple, Type, Union, cast
import warnings

from graphviz import Digraph, ExecutableNotFound
from graphviz import version as graphviz_version
from tqdm import tqdm
from tqdm.contrib.concurrent import process_map

from ford.sourceform import (
    ExternalFunction,
    ExternalInterface,
    ExternalModule,
    ExternalProgram,
    ExternalSourceFile,
    ExternalSubmodule,
    ExternalSubroutine,
    ExternalType,
    FortranBlockData,
    FortranContainer,
    FortranInterface,
    FortranModule,
    FortranProcedure,
    FortranProgram,
    FortranSourceFile,
    FortranSubmodule,
    FortranSubmoduleProcedure,
    FortranType,
)

try:
    graphviz_version()
    graphviz_installed = True
except ExecutableNotFound:
    graphviz_installed = False


HYPERLINK_RE = re.compile(
    r"^\s*<\s*a\s+.*href=(\"[^\"]+\"|'[^']+').*>(.*)</\s*a\s*>\s*$", re.IGNORECASE
)
WIDTH_RE = re.compile('width="(.*?)pt"', re.IGNORECASE)
HEIGHT_RE = re.compile('height="(.*?)pt"', re.IGNORECASE)
EM_RE = re.compile("<em>(.*)</em>", re.IGNORECASE)


def newdict(old, key, val):
    new = copy.copy(old)
    new[key] = val
    return new


def is_module(obj):
    return isinstance(obj, FortranModule)


def is_submodule(obj):
    return isinstance(obj, FortranSubmodule)


def is_type(obj):
    return isinstance(obj, FortranType)


def is_proc(obj):
    return isinstance(
        obj,
        (FortranProcedure, FortranInterface, FortranSubmoduleProcedure),
    )


def is_program(obj):
    return isinstance(obj, FortranProgram)


def is_sourcefile(obj):
    return isinstance(obj, FortranSourceFile)


def is_blockdata(obj):
    return isinstance(obj, FortranBlockData)


NodeCollection = Dict[FortranContainer, "BaseNode"]


class GraphData:
    """Stores graph nodes representing Fortran entities to be
    displayed in a graph, as well as some customisation options for
    graphs

    Parameters
    ----------
    parent_dir:
        Path to top of site
    coloured_edges:
        If true, arrows between nodes are coloured, otherwise they are black
    show_proc_parent:
        If true, the parent of a procedure is shown in the node label

    """

    def __init__(self, parent_dir: str, coloured_edges: bool, show_proc_parent: bool):
        self.submodules: NodeCollection = {}
        self.modules: NodeCollection = {}
        self.types: NodeCollection = {}
        self.procedures: NodeCollection = {}
        self.programs: NodeCollection = {}
        self.sourcefiles: NodeCollection = {}
        self.blockdata: NodeCollection = {}
        self.parent_dir = parent_dir
        self.coloured_edges = coloured_edges
        self.show_proc_parent = show_proc_parent

    def _get_collection_and_node_type(
        self, obj: FortranContainer
    ) -> Tuple[NodeCollection, Type["BaseNode"]]:
        """Helper function for `register` and `get_node`: get the
        appropriate container for ``obj``, and the corresponding node
        type

        """

        if is_submodule(obj):
            return self.submodules, SubmodNode
        if is_module(obj):
            return self.modules, ModNode
        if is_type(obj):
            return self.types, TypeNode
        if is_proc(obj):
            return self.procedures, ProcNode
        if is_program(obj):
            return self.programs, ProgNode
        if is_sourcefile(obj):
            return self.sourcefiles, FileNode
        if is_blockdata(obj):
            return self.blockdata, BlockNode

        raise BadType(
            f"Unrecognised object type '{type(obj).__name__}' for object '{obj}' when constructing graphs"
        )

    def register(
        self, obj: FortranContainer, hist: Optional[NodeCollection] = None
    ) -> None:
        """Create and store the graph node for ``obj``, if it hasn't
        already been registered

        Parameters
        ----------
        obj:
            Some Fortran entity
        hist:
            Collection of previously seen objects, used when
            registering children during node creation

        """

        collection, NodeType = self._get_collection_and_node_type(obj)
        if obj not in collection:
            collection[obj] = NodeType(obj, self, hist)

    def get_node(
        self, obj: FortranContainer, hist: Optional[NodeCollection] = None
    ) -> BaseNode:
        """Returns the node corresponding to ``obj``. If does not
        already exist then it will create it.

        Parameters
        ----------
        obj:
            Some Fortran entity
        hist:
            Collection of previously seen objects, used when
            registering children during node creation

        """
        hist = hist or {}

        if obj in hist:
            return hist[obj]

        collection, _ = self._get_collection_and_node_type(obj)
        if obj not in collection:
            self.register(obj, hist)

        return collection[obj]

    def get_module_node(self, mod: Union[FortranModule, str]) -> ModNode:
        if isinstance(mod, str):
            # Most likely a third-party module
            mod = ExternalModule(mod)
        return cast(ModNode, self.get_node(mod))

    def get_procedure_node(
        self,
        procedure: Union[FortranProcedure, str],
        hist: NodeCollection,
    ) -> ProcNode:
        if isinstance(procedure, str):
            # Most likely a third-party procedure
            procedure = ExternalSubroutine(procedure)
            procedure.proctype = "unknown"

        return cast(ProcNode, self.get_node(procedure, hist))

    def get_type_node(
        self, type_: Union[FortranType, str], hist: NodeCollection
    ) -> TypeNode:
        if isinstance(type_, str):
            # Most likely a third-party type
            type_ = ExternalType(type_)

        return cast(TypeNode, self.get_node(type_, hist))


class BaseNode:
    """Graph node representing some Fortran entity

    Parameters
    ----------
    obj:
        Fortran entity instance or name
    graph_data:
        Collection of nodes for other entities
    hist:

    """

    colour = "#777777"

    def __init__(
        self,
        obj: Union[FortranContainer, str],
        graph_data: GraphData,
        hist: Optional[NodeCollection] = None,
    ):
        self.attribs = {"color": self.colour, "fontcolor": "white", "style": "filled"}
        if isinstance(
            obj,
            (
                ExternalModule,
                ExternalSubmodule,
                ExternalType,
                ExternalSubroutine,
                ExternalFunction,
                ExternalInterface,
                ExternalSubroutine,
                ExternalProgram,
                ExternalSourceFile,
            ),
        ):
            self.fromstr = True
            obj = obj.name

        self.url = None
        if isinstance(obj, str):
            self.fromstr = True
            if m := HYPERLINK_RE.match(obj):
                self.url = m.group(1)[1:-1]
                self.name = m.group(2)
            else:
                self.name = obj
            self.ident = self.name
        else:
            self.fromstr = False
            d = obj.get_dir() or "none"
            self.ident = f"{d}~{obj.ident}"
            self.name = obj.name
            if m := EM_RE.search(self.name):
                self.name = "<<i>" + m.group(1).strip() + "</i>>"
            self.url = obj.get_url()

        self.attribs["label"] = self.name
        if self.url and getattr(obj, "visible", True):
            if self.fromstr or hasattr(obj, "external_url"):
                self.attribs["URL"] = self.url
            else:
                self.attribs["URL"] = graph_data.parent_dir + self.url
        self.afferent = 0
        self.efferent = 0

    def __eq__(self, other):
        return self.ident == other.ident

    def __lt__(self, other):
        return self.ident < other.ident

    def __hash__(self):
        # When making graphs in parallel, nodes might not have all
        # their attributes at some point?
        try:
            return hash(self.ident)
        except AttributeError:
            return id(self)


class ModNode(BaseNode):
    colour = "#337AB7"

    def __init__(self, obj, gd, hist=None):
        super().__init__(obj, gd)
        self.uses = set()
        self.used_by = set()
        self.children = set()
        if self.fromstr:
            return
        for u in obj.uses:
            n = gd.get_module_node(u)
            n.used_by.add(self)
            n.afferent += 1
            self.uses.add(n)
            self.efferent += n.efferent


class SubmodNode(ModNode):
    colour = "#5bc0de"

    def __init__(self, obj, gd, hist=None):
        super().__init__(obj, gd)
        del self.used_by
        if self.fromstr:
            return
        if obj.parent_submodule:
            self.ancestor = gd.get_node(obj.parent_submodule)
        else:
            self.ancestor = gd.get_module_node(obj.ancestor_module)
        self.ancestor.children.add(self)
        self.efferent += 1
        self.ancestor.afferent += 1


class TypeNode(BaseNode):
    colour = "#5cb85c"

    def __init__(self, obj, gd, hist=None):
        super().__init__(obj, gd)
        self.ancestor = None
        self.children = set()
        self.comp_types = dict()
        self.comp_of = dict()
        if self.fromstr:
            return

        hist = newdict(hist or {}, obj, self)

        if hasattr(obj, "external_url"):
            # Stop following chain, as this object is in an external project
            return

        if obj.extends:
            self.ancestor = gd.get_type_node(obj.extends, hist)
            self.ancestor.children.add(self)
            self.ancestor.visible = getattr(obj.extends, "visible", True)

        for var in obj.local_variables:
            if var.vartype not in ["type", "class"]:
                continue

            proto = var.proto[0]
            if proto == "*":
                continue

            node = gd.get_type_node(proto, hist)

            node.visible = getattr(proto, "visible", True)
            if self in node.comp_of:
                node.comp_of[self] += ", " + var.name
            else:
                node.comp_of[self] = var.name
            if node in self.comp_types:
                self.comp_types[node] += ", " + var.name
            else:
                self.comp_types[node] = var.name


class ProcNode(BaseNode):
    COLOURS = {"subroutine": "#d9534f", "function": "#d94e8f", "interface": "#A7506F"}

    @property
    def colour(self):
        return ProcNode.COLOURS.get(self.proctype.lower(), super().colour)

    def __init__(self, obj, gd, hist=None):
        # ToDo: Figure out appropriate way to handle interfaces to routines in submodules.
        self.proctype = getattr(obj, "proctype", "")
        super().__init__(obj, gd)

        if parent := getattr(obj, "parent", None):
            parent_label = f"{parent.name}::"
        else:
            parent_label = ""
        if binding := getattr(obj, "binding", None):
            binding_label = f"{binding.parent.name}%"
        else:
            binding_label = ""
        self.attribs["label"] = f"{parent_label}{binding_label}{self.name}"

        self.uses = set()
        self.calls = set()
        self.called_by = set()
        self.interfaces = set()
        self.interfaced_by = set()

        if self.fromstr:
            return

        hist = newdict(hist or {}, obj, self)

        for u in getattr(obj, "uses", []):
            n = gd.get_module_node(u)
            n.used_by.add(self)
            self.uses.add(n)

        for c in getattr(obj, "calls", []):
            if getattr(c, "visible", True):
                n = gd.get_procedure_node(c, hist)
                n.called_by.add(self)
                self.calls.add(n)

        if obj.proctype.lower() != "interface":
            return

        for m in getattr(obj, "modprocs", []):
            if m.procedure and getattr(m.procedure, "visible", True):
                n = gd.get_procedure_node(m.procedure, hist)
                n.interfaced_by.add(self)
                self.interfaces.add(n)

        if (
            hasattr(obj, "procedure")
            and obj.procedure.module
            and obj.procedure.module is not True
            and getattr(obj.procedure.module, "visible", True)
        ):
            n = gd.get_procedure_node(obj.procedure.module, hist)
            n.interfaced_by.add(self)
            self.interfaces.add(n)


class ProgNode(BaseNode):
    colour = "#f0ad4e"

    def __init__(self, obj, gd, hist=None):
        super().__init__(obj, gd)
        self.uses = set()
        self.calls = set()
        if self.fromstr:
            return
        for u in obj.uses:
            n = gd.get_module_node(u)
            n.used_by.add(self)
            self.uses.add(n)

        for c in obj.calls:
            if not getattr(c, "visible", False):
                continue
            n = gd.get_procedure_node(c, {})
            n.called_by.add(self)
            self.calls.add(n)


class BlockNode(BaseNode):
    colour = "#5cb85c"

    def __init__(self, obj, gd, hist=None):
        super().__init__(obj, gd)
        self.uses = set()
        if self.fromstr:
            return
        for u in obj.uses:
            n = gd.get_module_node(u)
            n.used_by.add(self)
            self.uses.add(n)


class FileNode(BaseNode):
    colour = "#f0ad4e"

    def __init__(self, obj, gd, hist=None):
        super().__init__(obj, gd)
        self.afferent = set()  # Things depending on this file
        self.efferent = set()  # Things this file depends on

        if self.fromstr:
            return

        hist = newdict(hist or {}, obj, self)

        for mod in itertools.chain(
            obj.modules,
            obj.submodules,
            obj.functions,
            obj.subroutines,
            obj.programs,
            obj.blockdata,
        ):
            for dep in mod.deplist:
                sourcefile = dep.hierarchy[0]
                if sourcefile == obj:
                    continue
                n = hist.get(sourcefile, gd.get_node(sourcefile, hist))
                n.afferent.add(self)
                self.efferent.add(n)


def _edge(
    tail: BaseNode, head: BaseNode, style: str, colour: str, label: Optional[str] = None
) -> Dict:
    return {
        "tail_node": tail,
        "head_node": head,
        "edge": {
            "tail_name": tail.ident,
            "head_name": head.ident,
            "style": style,
            "color": colour,
            "label": label,
        },
    }


def _solid_edge(
    tail: BaseNode, head: BaseNode, colour: str, label: Optional[str] = None
) -> Dict:
    return _edge(tail, head, "solid", colour, label)


def _dashed_edge(
    tail: BaseNode, head: BaseNode, colour: str, label: Optional[str] = None
) -> Dict:
    return _edge(tail, head, "dashed", colour, label)


if graphviz_installed:
    # Create the legends for the graphs. These are their own separate graphs,
    # without edges
    gd = GraphData("", False, False)

    # Graph nodes for a bunch of fake entities that we'll use in the legend
    _module = gd.get_node(ExternalModule("Module"))
    _submodule = gd.get_node(ExternalSubmodule("Submodule"))
    _type = gd.get_node(ExternalType("Type"))
    _subroutine = gd.get_node(ExternalSubroutine("Subroutine"))
    _function = gd.get_node(ExternalFunction("Function"))
    _interface = gd.get_node(ExternalInterface("Interface"))
    _unknown_proc = ExternalSubroutine("Unknown Procedure Type")
    _unknown_proc.proctype = "Unknown"
    _unknown = gd.get_node(_unknown_proc)
    _program = gd.get_node(ExternalProgram("Program"))
    _sourcefile = gd.get_node(ExternalSourceFile("Source File"))

    def _make_legend(entities):
        """Make a legend containing a collection of entities"""
        dot = Digraph(
            "Graph Key",
            graph_attr={"size": "8.90625,1000.0", "concentrate": "false"},
            node_attr={
                "shape": "box",
                "height": "0.0",
                "margin": "0.08",
                "fontname": "Helvetica",
                "fontsize": "10.5",
            },
            edge_attr={"fontname": "Helvetica", "fontsize": "9.5"},
            format="svg",
            engine="dot",
        )
        for entity in entities:
            dot.node(entity.name, **entity.attribs)
        dot.node("This Page's Entity")
        return dot.pipe().decode("utf-8")

    mod_svg = _make_legend([_module, _submodule, _subroutine, _function, _program])
    type_svg = _make_legend([_type])
    call_svg = _make_legend([_subroutine, _function, _interface, _unknown, _program])
    file_svg = _make_legend([_sourcefile])
else:
    mod_svg = ""
    type_svg = ""
    call_svg = ""
    file_svg = ""

NODE_DIAGRAM = "<p>Nodes of different colours represent the following: </p>"

MOD_GRAPH_KEY = f"""
{NODE_DIAGRAM}
{mod_svg}
<p>Solid arrows point from a submodule to the (sub)module which it is
descended from. Dashed arrows point from a module or program unit to 
modules which it uses.
</p>
"""  # noqa W291

TYPE_GRAPH_KEY = f"""
{NODE_DIAGRAM}
{type_svg}
<p>Solid arrows point from a derived type to the parent type which it
extends. Dashed arrows point from a derived type to the other
types it contains as a components, with a label listing the name(s) of
said component(s).
</p>
"""

CALL_GRAPH_KEY = f"""
{NODE_DIAGRAM}
{call_svg}
<p>Solid arrows point from a procedure to one which it calls. Dashed 
arrows point from an interface to procedures which implement that interface.
This could include the module procedures in a generic interface or the
implementation in a submodule of an interface in a parent module.
</p>
"""  # noqa W291

FILE_GRAPH_KEY = f"""
{NODE_DIAGRAM}
{file_svg}
<p>Solid arrows point from a file to a file which it depends on. A file
is dependent upon another if the latter must be compiled before the former
can be.
</p>
"""

COLOURED_NOTICE = """Where possible, edges connecting nodes are
given different colours to make them easier to distinguish in
large graphs."""

del call_svg
del file_svg
del type_svg
del mod_svg


class FortranGraph:
    """Graph of some relationship for a given entity

    Parameters
    ----------
    root:
        Top-level entity or entities in graph
    data:
        Collection of nodes and graph customisation options
    ident:
        Alternative identification for graph, and used as base name
        for saved files. If there are multiple entities in ``root``,
        and ``ident`` isn't given, it is set from the first entity in
        ``root``

    Attributes
    ----------
    hop_nodes:
        Nodes of the hop which exceed the maximum
    hop_edges:
        Edges of the hop which exceed the maximum
    added:
        Set of nodes in graph
    max_nesting:
        Maximum number of hops allowed. Set from maximum value of
        ``graph_maxdepth`` in ``root.meta``
    max_nodes:
        Maximum number of nodes allowed. Set from maximum value of
        ``graph_maxnodes`` in ``root.meta``
    warn:
        If true, show warnings if graphs exceed ``max_nesting`` or
        ``max_nodes``
    truncated:
        Nesting level where the graph was truncated
    """

    RANKDIR = "RL"
    _should_add_nested_nodes = False
    _legend = ""

    def __init__(
        self,
        root: Union[FortranContainer, Iterable[FortranContainer]],
        data: GraphData,
        ident: Optional[str] = None,
    ):
        self.root = []
        self.data = data
        self.hop_nodes: List[BaseNode] = []
        self.hop_edges: List[BaseNode] = []
        self.added: Set[BaseNode] = set()
        self.max_nesting = 0
        self.max_nodes = 1
        self.warn = False
        self.truncated = -1

        if not isinstance(root, Iterable):
            root = [root]
        root = sorted(list(root))

        for r in root:
            self.root.append(self.data.get_node(r))
            self.max_nesting = max(self.max_nesting, int(r.meta["graph_maxdepth"]))
            self.max_nodes = max(self.max_nodes, int(r.meta["graph_maxnodes"]))
            self.warn = self.warn or (r.settings["warn"])

        ident = ident or f"{root[0].get_dir()}~~{root[0].ident}"
        self.ident = f"{ident}~~{self.__class__.__name__}"
        self.imgfile = self.ident
        self.dot = Digraph(
            self.ident,
            graph_attr={
                "size": "8.90625,1000.0",
                "rankdir": self.RANKDIR,
                "concentrate": "true",
                "id": self.ident,
            },
            node_attr={
                "shape": "box",
                "height": "0.0",
                "margin": "0.08",
                "fontname": "Helvetica",
                "fontsize": "10.5",
            },
            edge_attr={"fontname": "Helvetica", "fontsize": "9.5"},
            format="svg",
            engine="dot",
        )
        # add root nodes to the graph
        for n in sorted(self.root):
            if len(self.root) == 1:
                self.dot.node(n.ident, label=n.attribs["label"])
            else:
                self.dot.node(n.ident, **n.attribs)
            self.added.add(n)
        # add nodes and edges depending on the root nodes to the graph
        self.add_nodes(self.root)

        if graphviz_installed:
            self.svg_src = self.dot.pipe().decode("utf-8")
            self.svg_src = self.svg_src.replace(
                "<svg ", '<svg id="' + re.sub(r"[^\w]", "", self.ident) + '" '
            )
            if match := WIDTH_RE.search(self.svg_src):
                width = int(match.group(1))
            else:
                width = 0
            if isinstance(self, (ModuleGraph, CallGraph, TypeGraph)):
                self.scaled = width >= 855
            else:
                self.scaled = width >= 641
        else:
            self.svg_src = ""
            self.scaled = False

    def add_to_graph(self, nodes, edges, nesting):
        """
        Adds nodes and edges to the graph as long as the maximum number
        of nodes is not exceeded.
        All edges are expected to have a reference to an entry in nodes.
        If the list of nodes is not added in the first hop due to graph
        size limitations, they are stored in hop_nodes.
        If the graph was extended the function returns True, otherwise the
        result will be False.
        """
        if (len(nodes) + len(self.added)) > self.max_nodes:
            if nesting < 2:
                self.hop_nodes = nodes
                self.hop_edges = edges
            self.truncated = nesting
            return False

        for n in sorted(nodes):
            strattribs = {key: str(a) for key, a in n.attribs.items()}
            self.dot.node(n.ident, **strattribs)
        for edge in edges:
            self.dot.edge(**edge["edge"])
        self.added.update(nodes)
        return True

    def __str__(self):
        """
        The string of the graph is its HTML representation.
        It will only be created if it is not too large.
        If the graph is overly large but can represented by a single node
        with many dependencies it will be shown as a table instead to ease
        the rendering in browsers.
        """

        graph_as_table = len(self.hop_nodes) > 0 and len(self.root) == 1

        # Do not render empty graphs
        if len(self.added) <= 1 and not graph_as_table:
            return ""

        # Do not render overly large graphs.
        if len(self.added) > self.max_nodes and self.warn:
            warnings.warn(
                f"Warning: Not showing graph {self.ident} as it would exceed the maximal number of {self.max_nodes} nodes"
            )
            return ""
        # Do not render incomplete graphs.
        if len(self.added) < len(self.root) and self.warn:
            warnings.warn(
                f"Warning: Not showing graph {self.ident} as it would be incomplete"
            )
            return ""

        if self.truncated > 0 and self.warn:
            warnings.warn(
                f"Warning: Graph {self.ident} is truncated after {self.truncated} hops"
            )

        if graph_as_table:
            rettext = self._make_graph_as_table()
        # generate svg graph
        else:
            rettext = f'<div class="depgraph">{self.svg_src}</div>'
            # add zoom ability for big graphs
            if self.scaled:
                zoomName = re.sub(r"[^\w]", "", self.ident)
                rettext += f"""\
                <script>
                  var pan{zoomName} = svgPanZoom('#{zoomName}',
                    {{zoomEnabled: true, controlIconsEnabled: true, fit: true, center: true,}}
                  );
                </script>"""

        legend_graph = f"""\
        <div><a type="button" class="graph-help" data-toggle="modal" href="#graph-help-text">Help</a></div>
          <div class="modal fade" id="graph-help-text" tabindex="-1" role="dialog">
            <div class="modal-dialog modal-lg" role="document">
              <div class="modal-content">
                <div class="modal-header">
                  <button type="button" class="close" data-dismiss="modal" aria-label="Close">
                    <span aria-hidden="true">&times;</span>
                  </button>
                  <h4 class="modal-title" id="-graph-help-label">Graph Key</h4>
                </div>
              <div class="modal-body">{self._legend} {COLOURED_NOTICE if self.data.coloured_edges else ""}</div>
            </div>
          </div>
        </div>"""
        return rettext + legend_graph

    def _make_graph_as_table(self):
        # generate a table graph if maximum number of nodes gets exceeded in
        # the first hop and there is only one root node.

        # Base templates for the arrows along edges
        arrow_shaft = '<td class="{style}{text_loc}">{label}</td>'
        arrow_right = f'{arrow_shaft}<td rowspan="2" class="triangle-right"></td>'
        arrow_left = f'<td rowspan="2" class="triangle-left"></td>{arrow_shaft}'

        # Work out if the root node is the head or tail of the
        # arrow, and which direction the arrows point in
        if self.hop_edges[0]["edge"]["tail_name"] == self.root[0].ident:
            key = "head_node"
            root_on_left = self.RANKDIR == "LR"
            arrowtemp = arrow_right if root_on_left else arrow_left
        else:
            key = "tail_node"
            root_on_left = self.RANKDIR == "RL"
            arrowtemp = arrow_left if root_on_left else arrow_right

        # Sort nodes in alphabetical order by either the head or
        # tail node's label
        self.hop_edges.sort(key=lambda x: x[key].attribs["label"].lower())

        # Now construct each node and associated edge as a single row in a table.
        # The root node takes up one column and spans all rows
        total_rows = len(self.hop_nodes) * 2 + 1
        root = f'<td class="root" rowspan="{total_rows}">{self.root[0].attribs["label"]}</td>'
        rows = ""
        for edge in self.hop_edges:
            style = edge["edge"]["style"]
            # The 'w' here is in white and is used to correctly position the arrow
            # shaft in the centre of the arrowhead
            label = edge["edge"]["label"] or "w"
            text_loc = "Bottom" if label == "w" else "Text"
            arrow_args = {"style": style, "label": label, "text_loc": text_loc}
            arrow = arrowtemp.format(**arrow_args)
            attribs = edge[key].attribs
            try:
                link = f'<a href="{attribs["URL"]}">{attribs["label"]}</a></td>'
            except KeyError:
                link = f'{attribs["label"]}</td>'

            node = f'<td rowspan="2" class="node" bgcolor="{attribs["color"]}">{link}'

            root_arrow = (
                f"{root}{arrow}{node}" if root_on_left else f"{node}{arrow}{root}"
            )
            rows += f'<tr>{root_arrow}</tr>\n<tr><td class="{style}Top">w</td></tr>\n'
            # Root node is handled by first row in table, so clear it
            root = ""

        return f'<table class="graph">\n{rows}</table>\n'

    def __nonzero__(self):
        return self.__bool__()

    def __bool__(self):
        return bool(self.__str__())

    def create_svg(self, out_location: pathlib.Path):
        if len(self.added) > len(self.root):
            out_location = pathlib.Path(out_location)
            self._create_image_file(out_location / self.imgfile)

    def _create_image_file(self, filename: pathlib.Path):
        if not graphviz_installed:
            return

        self.dot.render(str(filename), cleanup=False)
        filename.rename(str(filename) + ".gv")

    def add_nodes(self, nodes, nesting=1):
        """Add nodes and edges to this graph, based on the collection ``nodes``

        Subclasses should implement `_add_node`, and optionally
        `_extra_attributes`

        """
        hop_nodes = set()  # nodes in this hop
        hop_edges = []  # edges in this hop

        total_len = len(nodes)

        def rainbowcolour(depth, maxd):
            if not self.data.coloured_edges:
                return "#000000"
            (r, g, b) = colorsys.hsv_to_rgb(float(depth) / maxd, 1.0, 1.0)
            return f"#{int(255 * r)}{int(255 * g)}{int(255 * b)}"

        for i, node in enumerate(sorted(nodes)):
            colour = rainbowcolour(i, total_len)

            self._add_node(hop_nodes, hop_edges, node, colour)

        if not self.add_to_graph(hop_nodes, hop_edges, nesting):
            return

        self._extra_attributes()

        if self._should_add_nested_nodes:
            self._add_nested_nodes(hop_nodes, nesting)

    def _add_nested_nodes(self, hop_nodes, nesting):
        """Handles nested nodes"""
        if len(hop_nodes) == 0:
            return

        if nesting < self.max_nesting:
            self.add_nodes(hop_nodes, nesting=nesting + 1)
        else:
            self.truncated = nesting

    def _add_node(self, hop_nodes, hop_edges, node, colour):
        """Add a single node and its edges to this graph, typically by
        iterating over parents/children

        """

        raise NotImplementedError

    def _extra_attributes(self):
        """Add any extra attributes to the graph"""
        pass


class ModuleGraph(FortranGraph):
    """Shows the relationship between modules and submodules"""

    _legend = MOD_GRAPH_KEY

    def _add_node(self, hop_nodes, hop_edges, node, colour):
        for nu in sorted(node.uses):
            if nu not in self.added:
                hop_nodes.add(nu)
            hop_edges.append(_dashed_edge(node, nu, colour))

        if hasattr(node, "ancestor"):
            if node.ancestor not in self.added:
                hop_nodes.add(node.ancestor)
            hop_edges.append(_solid_edge(node, node.ancestor, colour))

    def _extra_attributes(self):
        self.dot.attr("graph", size="11.875,1000.0")


class UsesGraph(FortranGraph):
    """Graphs how modules use other modules, including ancestor (sub)modules"""

    _should_add_nested_nodes = True
    _legend = MOD_GRAPH_KEY

    def _add_node(self, hop_nodes, hop_edges, node, colour):
        for nu in sorted(node.uses):
            if nu not in self.added:
                hop_nodes.add(nu)
            hop_edges.append(_dashed_edge(node, nu, colour))

        if hasattr(node, "ancestor"):
            if node.ancestor not in self.added:
                hop_nodes.add(node.ancestor)
            hop_edges.append(_solid_edge(node, node.ancestor, colour))


class UsedByGraph(FortranGraph):
    """Graphs how modules are used by other modules"""

    _should_add_nested_nodes = True
    _legend = MOD_GRAPH_KEY

    def _add_node(self, hop_nodes, hop_edges, node, colour):
        for nu in sorted(getattr(node, "used_by", [])):
            if nu not in self.added:
                hop_nodes.add(nu)
            hop_edges.append(_dashed_edge(nu, node, colour))
        for c in sorted(getattr(node, "children", [])):
            if c not in self.added:
                hop_nodes.add(c)
            hop_edges.append(_solid_edge(c, node, colour))


class FileGraph(FortranGraph):
    """Graphs relationships between source files"""

    _legend = FILE_GRAPH_KEY

    def _add_node(self, hop_nodes, hop_edges, node, colour):
        for ne in sorted(node.efferent):
            if ne not in self.added:
                hop_nodes.add(ne)
            hop_edges.append(_solid_edge(ne, node, colour))


class EfferentGraph(FortranGraph):
    """Shows the relationship between the files which this one depends on"""

    _should_add_nested_nodes = True
    _legend = FILE_GRAPH_KEY

    def _add_node(self, hop_nodes, hop_edges, node, colour):
        for ne in sorted(node.efferent):
            if ne not in self.added:
                hop_nodes.add(ne)
            hop_edges.append(_dashed_edge(node, ne, colour))


class AfferentGraph(FortranGraph):
    """Shows the relationship between files which depend upon this one"""

    _should_add_nested_nodes = True
    _legend = FILE_GRAPH_KEY

    def _add_node(self, hop_nodes, hop_edges, node, colour):
        for na in sorted(node.afferent):
            if na not in self.added:
                hop_nodes.add(na)
            hop_edges.append(_dashed_edge(na, node, colour))


class TypeGraph(FortranGraph):
    """Graphs inheritance and composition relationships between derived types"""

    _legend = TYPE_GRAPH_KEY

    def _add_node(self, hop_nodes, hop_edges, node, colour):
        for keys in node.comp_types.keys():
            if keys not in self.added:
                hop_nodes.add(keys)
        for c in node.comp_types:
            if c not in self.added:
                hop_nodes.add(c)
            hop_edges.append(_dashed_edge(node, c, colour, node.comp_types[c]))
        if node.ancestor:
            if node.ancestor not in self.added:
                hop_nodes.add(node.ancestor)
            hop_edges.append(_solid_edge(node, node.ancestor, colour))

    def _extra_attributes(self):
        self.dot.attr("graph", size="11.875,1000.0")


class InheritsGraph(FortranGraph):
    """Graphs types that this type inherits from"""

    _should_add_nested_nodes = True
    _legend = TYPE_GRAPH_KEY

    def _add_node(self, hop_nodes, hop_edges, node, colour):
        for c in node.comp_types:
            if c not in self.added:
                hop_nodes.add(c)
            hop_edges.append(_dashed_edge(node, c, colour, node.comp_types[c]))
        if node.ancestor:
            if node.ancestor not in self.added:
                hop_nodes.add(node.ancestor)
            hop_edges.append(_solid_edge(node, node.ancestor, colour))


class InheritedByGraph(FortranGraph):
    """Graphs types that inherit this type"""

    _should_add_nested_nodes = True
    _legend = TYPE_GRAPH_KEY

    def _add_node(self, hop_nodes, hop_edges, node, colour):
        for c in node.comp_of:
            if c not in self.added:
                hop_nodes.add(c)
            hop_edges.append(_dashed_edge(c, node, colour, node.comp_of[c]))
        for c in node.children:
            if c not in self.added:
                hop_nodes.add(c)
            hop_edges.append(_solid_edge(c, node, colour))


class CallGraph(FortranGraph):
    """
    Adds edges indicating the call-tree for the procedures listed in
    the nodes.
    """

    RANKDIR = "LR"
    _legend = CALL_GRAPH_KEY

    def _add_node(self, hop_nodes, hop_edges, node, colour):
        for p in sorted(node.calls):
            if p not in hop_nodes:
                hop_nodes.add(p)
            hop_edges.append(_solid_edge(node, p, colour))
        for p in sorted(getattr(node, "interfaces", [])):
            if p not in hop_nodes:
                hop_nodes.add(p)
            hop_edges.append(_dashed_edge(node, p, colour))

    def _extra_attributes(self):
        self.dot.attr("graph", size="11.875,1000.0")
        self.dot.attr("graph", concentrate="false")


class CallsGraph(FortranGraph):
    """Graphs procedures that this procedure calls"""

    RANKDIR = "LR"
    _should_add_nested_nodes = True
    _legend = CALL_GRAPH_KEY

    def _add_node(self, hop_nodes, hop_edges, node, colour):
        for p in sorted(node.calls):
            if p not in self.added:
                hop_nodes.add(p)
            hop_edges.append(_solid_edge(node, p, colour))
        for p in sorted(getattr(node, "interfaces", [])):
            if p not in self.added:
                hop_nodes.add(p)
            hop_edges.append(_dashed_edge(node, p, colour))

    def _extra_attributes(self):
        self.dot.attr("graph", concentrate="false")


class CalledByGraph(FortranGraph):
    """Graphs procedures called by this procedure"""

    RANKDIR = "LR"
    _should_add_nested_nodes = True
    _legend = CALL_GRAPH_KEY

    def _add_node(self, hop_nodes, hop_edges, node, colour):
        if isinstance(node, ProgNode):
            return
        for p in sorted(node.called_by):
            if p not in self.added:
                hop_nodes.add(p)
            hop_edges.append(_solid_edge(p, node, colour))
        for p in sorted(getattr(node, "interfaced_by", [])):
            if p not in self.added:
                hop_nodes.add(p)
            hop_edges.append(_dashed_edge(p, node, colour))

    def _extra_attributes(self):
        self.dot.attr("graph", concentrate="false")


class BadType(Exception):
    """
    Raised when a type is passed to GraphData.register() which is not
    accepted.
    """

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


def outputFuncWrap(args):
    """Wrapper function for output graphs -- needed to allow multiprocessing to
    pickle the function (must be at top level)"""

    for f in args[0:-1]:
        f.create_svg(args[-1])

    return None


class GraphManager:
    """Collection of graphs of the various relationships between a set
    of entities

    Contains graphs of module use relations, type relations, call
    trees, etc. It manages these, ensures that everything that is
    needed is added at the correct time, and produces the plots for
    the list pages.

    Parameters
    ----------
    base_url:
        The URL at which the documentation will be stored. If using
        relative URLs then should be '..'.
    outdir:
        The directory in which the documentation will be produced.
    graphdir:
        The location of the graphs within the output tree.
    parentdir:
        Location of top-level directory
    coloured_edges:
        If true, arrows in graphs use different colours to help
        distinguish them
    show_proc_parent:
        If true, show the parent of a procedure in the call graph
        as part of the label
    save_graphs:
        If true, save graphs as separate files, as well as embedding
        them in the HTML
    """

    def __init__(
        self,
        base_url: os.PathLike,
        outdir: os.PathLike,
        graphdir: os.PathLike,
        parentdir: str,
        coloured_edges: bool,
        show_proc_parent: bool,
        save_graphs: bool = False,
    ):
        self.graph_objs: List[FortranContainer] = []
        self.modules: Set[FortranContainer] = set()
        self.programs: Set[FortranContainer] = set()
        self.procedures: Set[FortranContainer] = set()
        self.types: Set[FortranContainer] = set()
        self.sourcefiles: Set[FortranContainer] = set()
        self.blockdata: Set[FortranContainer] = set()
        self.save_graphs = save_graphs
        self.graphdir = pathlib.Path(graphdir)
        self.usegraph = None
        self.typegraph = None
        self.callgraph = None
        self.filegraph = None
        self.data = GraphData(parentdir, coloured_edges, show_proc_parent)

    def register(self, obj: FortranContainer):
        """Register ``obj`` as a node to be used in graphs"""
        if obj.meta["graph"]:
            self.data.register(obj)
            self.graph_objs.append(obj)

    def graph_all(self):
        """Create all graphs"""
        for obj in tqdm(sorted(self.graph_objs), unit="", desc="Generating graphs"):
            if is_module(obj):
                obj.usesgraph = UsesGraph(obj, self.data)
                obj.usedbygraph = UsedByGraph(obj, self.data)
                self.modules.add(obj)
            elif is_type(obj):
                obj.inhergraph = InheritsGraph(obj, self.data)
                obj.inherbygraph = InheritedByGraph(obj, self.data)
                self.types.add(obj)
            elif is_proc(obj):
                obj.callsgraph = CallsGraph(obj, self.data)
                obj.calledbygraph = CalledByGraph(obj, self.data)
                obj.usesgraph = UsesGraph(obj, self.data)
                self.procedures.add(obj)
            elif is_program(obj):
                obj.usesgraph = UsesGraph(obj, self.data)
                obj.callsgraph = CallsGraph(obj, self.data)
                self.programs.add(obj)
            elif is_sourcefile(obj):
                obj.afferentgraph = AfferentGraph(obj, self.data)
                obj.efferentgraph = EfferentGraph(obj, self.data)
                self.sourcefiles.add(obj)
            elif is_blockdata(obj):
                obj.usesgraph = UsesGraph(obj, self.data)
                self.blockdata.add(obj)

        usenodes = sorted(list(self.modules))
        callnodes = sorted(list(self.procedures))
        for p in sorted(self.programs):
            if len(p.usesgraph.added) > 1:
                usenodes.append(p)
            if len(p.callsgraph.added) > 1:
                callnodes.append(p)
        for p in sorted(self.procedures):
            if len(p.usesgraph.added) > 1:
                usenodes.append(p)
        for b in self.blockdata:
            if len(b.usesgraph.added) > 1:
                usenodes.append(b)
        self.usegraph = ModuleGraph(usenodes, self.data, "module~~graph")
        self.typegraph = TypeGraph(self.types, self.data, "type~~graph")
        self.callgraph = CallGraph(callnodes, self.data, "call~~graph")
        self.filegraph = FileGraph(self.sourcefiles, self.data, "file~~graph")

    def output_graphs(self, njobs=0):
        """Save graphs to file"""

        if not self.save_graphs:
            return

        self.graphdir.mkdir(exist_ok=True, parents=True, mode=0o755)

        if njobs == 0:
            for m in self.modules:
                m.usesgraph.create_svg(self.graphdir)
                m.usedbygraph.create_svg(self.graphdir)
            for t in self.types:
                t.inhergraph.create_svg(self.graphdir)
                t.inherbygraph.create_svg(self.graphdir)
            for p in self.procedures:
                p.callsgraph.create_svg(self.graphdir)
                p.calledbygraph.create_svg(self.graphdir)
            for p in self.programs:
                p.callsgraph.create_svg(self.graphdir)
                p.usesgraph.create_svg(self.graphdir)
            for f in self.sourcefiles:
                f.afferentgraph.create_svg(self.graphdir)
                f.efferentgraph.create_svg(self.graphdir)
            for b in self.blockdata:
                b.usesgraph.create_svg(self.graphdir)
        else:
            args = []
            # Note we generate all graphs for a given object in one wrapper call
            # this is to try to ensure we don't get name collisions not present
            # in the serial version (e.g. due to calling usesgraph and usedbygraph on
            # a particular module in two different processes). May not actually be needed
            # commented block above allows testing of one graph per call approach.
            args.extend(
                [(m.usesgraph, m.usedbygraph, self.graphdir) for m in self.modules]
            )
            args.extend(
                [(m.inhergraph, m.inherbygraph, self.graphdir) for m in self.types]
            )
            args.extend(
                [
                    (m.callsgraph, m.calledbygraph, self.graphdir)
                    for m in self.procedures
                ]
            )
            args.extend(
                [(m.callsgraph, m.usesgraph, self.graphdir) for m in self.programs]
            )
            args.extend(
                [
                    (m.afferentgraph, m.efferentgraph, self.graphdir)
                    for m in self.sourcefiles
                ]
            )
            args.extend([(m.usesgraph, self.graphdir) for m in self.blockdata])

            process_map(
                outputFuncWrap,
                args,
                max_workers=njobs,
                desc="Writing graphs",
            )

        for graph in [self.usegraph, self.typegraph, self.callgraph, self.filegraph]:
            if graph:
                graph.create_svg(self.graphdir)



import shutil
import sys
import os
import pathlib
import copy
from urllib.parse import urlparse
import json

import ford

from bs4 import BeautifulSoup
import pytest


REMOTE_MODULES_JSON = [
    {
        "name": "remote_module",
        "external_url": "./module/remote_module.html",
        "obj": "module",
        "pub_procs": {
            "remote_sub": {
                "name": "remote_sub",
                "external_url": "./proc/remote_sub.html",
                "obj": "proc",
                "proctype": "Subroutine",
                "functions": [],
                "subroutines": [],
                "interfaces": [],
                "absinterfaces": [],
                "types": [],
                "variables": [],
            },
        },
        "pub_absints": {},
        "pub_types": {},
        "pub_vars": {},
        "functions": [],
        "subroutines": [
            {
                "name": "remote_sub",
                "external_url": "./proc/remote_sub.html",
                "obj": "proc",
                "proctype": "Subroutine",
                "functions": [],
                "subroutines": [],
                "interfaces": [],
                "absinterfaces": [],
                "types": [],
                "variables": [],
            }
        ],
        "interfaces": [],
        "absinterfaces": [],
        "types": [],
        "variables": [],
    }
]


class MockResponse:
    @staticmethod
    def read():
        return json.dumps(REMOTE_MODULES_JSON).encode("utf-8")


@pytest.fixture(scope="module")
def monkeymodule(request):
    """pytest won't let us use function-scope fixtures in module-scope
    fixtures, so we need to reimplement this with module scope"""
    mpatch = pytest.MonkeyPatch()
    yield mpatch
    mpatch.undo()


@pytest.fixture(scope="module")
def restore_macros_module():
    """pytest won't let us use function-scope fixtures in module-scope
    fixtures, so we need to reimplement this with module scope"""

    old_macros = copy.copy(ford.utils._MACRO_DICT)
    yield
    ford.utils._MACRO_DICT = copy.copy(old_macros)


@pytest.fixture(scope="module")
def external_project(tmp_path_factory, monkeymodule, restore_macros_module):
    """Generate the documentation for an "external" project and then
    for a "top level" one that uses the first.

    A remote external project is simulated through a mocked `urlopen`
    which returns `REMOTE_MODULES_JSON`

    """

    this_dir = pathlib.Path(__file__).parent
    path = tmp_path_factory.getbasetemp() / "external_project"
    shutil.copytree(this_dir / "../../test_data/external_project", path)

    external_project = path / "external_project"
    top_level_project = path / "top_level_project"

    # Run FORD in the two projects
    # First project has "externalize: True" and will generate JSON dump
    with monkeymodule.context() as m:
        os.chdir(external_project)
        m.setattr(sys, "argv", ["ford", "doc.md"])
        ford.run()

    def mock_open(*args, **kwargs):
        return MockResponse()

    # Second project uses JSON from first to link to external modules
    with monkeymodule.context() as m:
        os.chdir(top_level_project)
        m.setattr(sys, "argv", ["ford", "doc.md"])
        m.setattr(ford.utils, "urlopen", mock_open)
        ford.run()

    # Make sure we're in a directory where relative paths won't
    # resolve correctly
    os.chdir("/")

    return top_level_project, external_project


def test_external_project(external_project):
    """Check that we can build external projects and get the links correct"""

    top_level_project, _ = external_project

    # Read generated HTML
    with open(top_level_project / "doc/program/top_level.html", "r") as f:
        top_program_html = BeautifulSoup(f.read(), features="html.parser")

    # Find links to external modules
    uses_box = top_program_html.find(string="Uses").parent.parent.parent
    links = {tag.text: tag.a["href"] for tag in uses_box("li", class_=None)}

    assert len(links) == 3
    assert "external_module" in links
    local_url = urlparse(links["external_module"])
    assert pathlib.Path(local_url.path).is_file()

    assert "remote_module" in links
    remote_url = urlparse(links["remote_module"])
    assert remote_url.scheme == "https"


def test_procedure_module_use_links_(external_project):
    """Check that links to external modules used by functions are correct"""

    top_level_project, _ = external_project

    # Read generated HTML
    with open(top_level_project / "doc/proc/abortcriteria_load.html", "r") as f:
        procedure_html = BeautifulSoup(f.read(), features="html.parser")

    # Find links to external modules
    uses_box = procedure_html.find(string="Uses").parent.parent.parent
    links = {tag.text: tag.a["href"] for tag in uses_box("li", class_=None)}

    assert len(links) == 1
    assert "external_module" in links
    local_url = urlparse(links["external_module"])
    assert pathlib.Path(local_url.path).is_file()



# -*- coding: utf-8 -*-
#
#  project.py
#  This file is part of FORD.
#
#  Copyright 2014 Christopher MacMackin <cmacmackin@gmail.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

import os
import toposort
from itertools import chain
from typing import List

import ford.utils
import ford.sourceform
from ford.sourceform import (
    FortranCodeUnit,
    FortranModule,
    FortranSubmodule,
    ExternalModule,
)


INTRINSIC_MODS = {
    "iso_fortran_env": '<a href="http://fortranwiki.org/fortran/show/iso_fortran_env">iso_fortran_env</a>',
    "iso_c_binding": '<a href="http://fortranwiki.org/fortran/show/iso_c_binding">iso_c_binding</a>',
    "ieee_arithmetic": '<a href="http://fortranwiki.org/fortran/show/ieee_arithmetic">ieee_arithmetic</a>',
    "ieee_exceptions": '<a href="http://fortranwiki.org/fortran/show/IEEE+arithmetic">ieee_exceptions</a>',
    "ieee_features": '<a href="http://fortranwiki.org/fortran/show/IEEE+arithmetic">ieee_features</a>',
    "openacc": '<a href="https://www.openacc.org/sites/default/files/inline-images/Specification/OpenACC.3.0.pdf#page=85">openacc</a>',
    "omp_lib": '<a href="https://www.openmp.org/spec-html/5.1/openmpch3.html#x156-1890003">omp_lib</a>',
    "mpi": '<a href="http://www.mpi-forum.org/docs/mpi-3.1/mpi31-report/node410.htm">mpi</a>',
    "mpi_f08": '<a href="http://www.mpi-forum.org/docs/mpi-3.1/mpi31-report/node409.htm">mpi_f08</a>',
}


class Project(object):
    """
    An object which collects and contains all of the information about the
    project which is to be documented.
    """

    def __init__(self, settings):
        self.settings = settings
        self.name = settings["project"]
        self.external = settings["external"]
        self.topdirs = settings["src_dir"]
        self.extensions = settings["extensions"]
        self.fixed_extensions = settings["fixed_extensions"]
        self.extra_filetypes = settings["extra_filetypes"]
        self.display = settings["display"]
        self.encoding = settings["encoding"]

        html_incl_src = settings.get("incl_src", True)

        self.files = []
        self.modules = []
        self.programs = []
        self.procedures = []
        self.absinterfaces = []
        self.types = []
        self.submodules = []
        self.submodprocedures = []
        self.extra_files = []
        self.blockdata = []
        self.common = {}
        self.extModules = []
        self.extProcedures = []
        self.extInterfaces = []
        self.extTypes = []
        self.extVariables = []

        # Get all files within topdir, recursively
        srcdir_list = self.make_srcdir_list(settings["exclude_dir"])
        for curdir in srcdir_list:
            for item in [f for f in curdir.iterdir() if f.is_file()]:
                if item.name in settings["exclude"]:
                    continue

                filename = curdir / item
                relative_path = os.path.relpath(filename)
                extension = str(item.suffix)[1:]  # Don't include the initial '.'
                if extension in self.extensions or extension in self.fixed_extensions:
                    # Get contents of the file
                    print(f"Reading file {relative_path}")
                    if extension in settings["fpp_extensions"]:
                        preprocessor = settings["preprocessor"]
                    else:
                        preprocessor = None
                    try:
                        self.files.append(
                            ford.sourceform.FortranSourceFile(
                                str(filename),
                                settings,
                                preprocessor,
                                extension in self.fixed_extensions,
                                incl_src=html_incl_src,
                                encoding=self.encoding,
                            )
                        )
                    except Exception as e:
                        if not settings["dbg"]:
                            raise e

                        print(f"Warning: Error parsing {relative_path}.\n\t{e.args[0]}")
                        continue
                    for module in self.files[-1].modules:
                        self.modules.append(module)
                    for submod in self.files[-1].submodules:
                        self.submodules.append(submod)
                    for function in self.files[-1].functions:
                        function.visible = True
                        self.procedures.append(function)
                    for subroutine in self.files[-1].subroutines:
                        subroutine.visible = True
                        self.procedures.append(subroutine)
                    for program in self.files[-1].programs:
                        program.visible = True
                        self.programs.append(program)
                    for block in self.files[-1].blockdata:
                        self.blockdata.append(block)
                elif extension in self.extra_filetypes:
                    print(f"Reading file {relative_path}")
                    try:
                        self.extra_files.append(
                            ford.sourceform.GenericSource(str(filename), settings)
                        )
                    except Exception as e:
                        if not settings["dbg"]:
                            raise e

                        print(f"Warning: Error parsing {relative_path}.\n\t{e.args[0]}")
                        continue

    def warn(self, message):
        if self.settings["warn"]:
            print(f"Warning: {message}")

    @property
    def allfiles(self):
        """Instead of duplicating files, it is much more efficient to create the itterator on the fly"""
        for f in self.files:
            yield f
        for f in self.extra_files:
            yield f

    def __str__(self):
        return self.name

    def correlate(self):
        """
        Associates various constructs with each other.
        """

        print("Correlating information from different parts of your project...")

        non_local_mods = INTRINSIC_MODS.copy()
        for item in self.settings["extra_mods"]:
            if not item:
                continue
            try:
                name, url = item.split(":", 1)
            except ValueError:
                raise ValueError(
                    f"Could not parse 'extra_mods' item in project settings: '{item}'\n"
                    "Expected something of the form 'module_name: http://link.to/module'"
                )
            name = name.strip()
            url = url.strip().strip(r"\"'").strip()
            non_local_mods[name.lower()] = f'<a href="{url}">{name}</a>'

        self.extModules.extend(
            [ExternalModule(name, url) for name, url in non_local_mods.items()]
        )

        # load external FORD FortranModules
        ford.utils.external(self)

        # Match USE statements up with the module objects or links
        for entity in chain(
            self.modules,
            self.procedures,
            self.programs,
            self.submodules,
            self.blockdata,
        ):
            find_used_modules(entity, self.modules, self.submodules, self.extModules)

        def get_deps(item):
            uselist = [m[0] for m in item.uses]
            for procedure in item.routines:
                uselist.extend(get_deps(procedure))
            return uselist

        def filter_modules(entity) -> List[FortranModule]:
            """Return a list of `FortranModule` from the dependencies of `entity`"""
            return [dep for dep in get_deps(entity) if type(dep) is FortranModule]

        # Get the order to process other correlations with
        for mod in self.modules:
            mod.deplist = filter_modules(mod)

        for mod in self.submodules:
            if type(mod.ancestor_module) is not FortranModule:
                self.warn(
                    f"Could not identify ancestor MODULE of SUBMODULE '{mod.name}'. "
                )
                continue

            if not isinstance(mod.parent_submodule, (FortranSubmodule, type(None))):
                self.warn(
                    f"Could not identify parent SUBMODULE of SUBMODULE '{mod.name}'"
                )

            mod.deplist = [
                mod.parent_submodule or mod.ancestor_module
            ] + filter_modules(mod)

        deplist = {
            module: set(module.deplist)
            for module in chain(self.modules, self.submodules)
        }

        # Get dependencies for programs and top-level procedures as well,
        # if dependency graphs are to be produced
        if self.settings["graph"]:
            for entity in chain(self.procedures, self.programs, self.blockdata):
                entity.deplist = set(filter_modules(entity))

        ranklist = toposort.toposort_flatten(deplist)
        for proc in self.procedures:
            if proc.parobj == "sourcefile":
                ranklist.append(proc)
        ranklist.extend(self.programs)
        ranklist.extend(self.blockdata)

        # Perform remaining correlations for the project
        for container in ranklist:
            if type(container) != str:
                container.correlate(self)
        for container in ranklist:
            if type(container) != str:
                container.prune()

        if self.settings["project_url"] == ".":
            url = ".."
        else:
            url = self.settings["project_url"]

        # Mapping of various entity containers in code units to the
        # corresponding container in the project
        CONTAINERS = {
            "functions": "procedures",
            "subroutines": "procedures",
            "interfaces": "procedures",
            "absinterfaces": "absinterfaces",
            "types": "types",
            "modfunctions": "submodprocedures",
            "modsubroutines": "submodprocedures",
            "modprocedures": "submodprocedures",
        }

        # Gather all the entity containers from each code unit in each
        # file into the corresponding project container
        for sfile in self.files:
            for code_unit in chain(
                sfile.modules, sfile.submodules, sfile.programs, sfile.blockdata
            ):
                for entity_kind, container in CONTAINERS.items():
                    entities = getattr(code_unit, entity_kind, [])
                    getattr(self, container).extend(entities)

        def sum_lines(*argv, **kwargs):
            """Wrapper for minimizing memory consumption"""
            routine = kwargs.get("func", "num_lines")
            n = 0
            for arg in argv:
                for item in arg:
                    n += getattr(item, routine)
            return n

        self.mod_lines = sum_lines(self.modules, self.submodules)
        self.proc_lines = sum_lines(self.procedures)
        self.file_lines = sum_lines(self.files)
        self.type_lines = sum_lines(self.types)
        self.type_lines_all = sum_lines(self.types, func="num_lines_all")
        self.absint_lines = sum_lines(self.absinterfaces)
        self.prog_lines = sum_lines(self.programs)
        self.block_lines = sum_lines(self.blockdata)
        print()

    def markdown(self, md, base_url=".."):
        """
        Process the documentation with Markdown to produce HTML.
        """
        print("\nProcessing documentation comments...")
        ford.sourceform.set_base_url(base_url)
        if self.settings["warn"]:
            print()
        for src in self.allfiles:
            src.markdown(md, self)

    def make_links(self, base_url=".."):
        """
        Substitute intrasite links to documentation for other parts of
        the program.
        """
        ford.sourceform.set_base_url(base_url)
        for src in self.allfiles:
            src.make_links(self)

    def make_srcdir_list(self, exclude_dirs):
        """
        Like `os.walk`, except that:

        a) directories listed in exclude_dir are excluded with all
           their subdirectories
        b) absolute paths are returned
        """
        srcdir_list = []
        for topdir in self.topdirs:
            srcdir_list.append(topdir)
            srcdir_list += self.recursive_dir_list(topdir, exclude_dirs)
        return srcdir_list

    def recursive_dir_list(self, topdir, skip):
        dir_list = []
        for entry in os.listdir(topdir):
            abs_entry = ford.utils.normalise_path(topdir, entry)
            if abs_entry.is_dir() and (abs_entry not in skip):
                dir_list.append(abs_entry)
                dir_list += self.recursive_dir_list(abs_entry, skip)
        return dir_list


def find_used_modules(
    entity: FortranCodeUnit,
    modules: List[FortranModule],
    submodules: List[FortranSubmodule],
    external_modules: List[ExternalModule],
) -> None:
    """Find the module objects (or links to intrinsic/external
    module) for all of the ``USED``d names in ``entity``

    Parameters
    ----------
    entity
        A program, module, submodule or procedure
    modules
        Known Fortran modules
    intrinsic_modules
        Known intrinsic Fortran modules
    submodules
        Known Fortran submodules
    external_modules
        Known external Fortran modules

    """
    # Find the modules that this entity uses
    for dependency in entity.uses:
        dependency_name = dependency[0].lower()
        for candidate in chain(modules, external_modules):
            if dependency_name == candidate.name.lower():
                dependency[0] = candidate
                break

    # Find the ancestor of this submodule (if entity is one)
    if getattr(entity, "parent_submodule", None):
        parent_submodule_name = entity.parent_submodule.lower()
        for submod in submodules:
            if parent_submodule_name == submod.name.lower():
                entity.parent_submodule = submod
                break

    if hasattr(entity, "ancestor_module"):
        ancestor_module_name = entity.ancestor_module.lower()
        for mod in modules:
            if ancestor_module_name == mod.name.lower():
                entity.ancestor_module = mod
                break

    # Find the modules that this entity's procedures use
    for procedure in entity.routines:
        find_used_modules(procedure, modules, submodules, external_modules)
